/*
 * To change this license header, choose License Headers in Project Properties.
 * To change this template file, choose Tools | Templates
 * and open the template in the editor.
 */

#include <stdlib.h>
#include <string.h>
#include <math.h>
#include <samplerate.h>
#include <sndfile.h>

#include "resampled.h"


#define	BUFFER_LEN 8192

// the same as enum srate
static const int srate_values[SRATE_COUNT] =
{
    8000, 11025, 16000, 22050, 32000, 44100, 48000, 96000, 192000
};

static void resampled_data_init(struct resampled_data *data)
{
    memset(data, 0, sizeof(struct resampled_data));
}

static void resampled_data_free(struct resampled_data *data)
{
    if (data == NULL)
        return;
    
    if (data->samples != NULL)
    {
        free(data->samples);
        data->samples = NULL;
    }
}

struct resampled_data *resampled_get(resampled_array_t arr, unsigned int rate)
{
    int i;

    if (arr == NULL)
        return NULL;
    
    for (i = 0; i < SRATE_COUNT; i++)
        if (srate_values[i] == rate)
            return &arr[i];
    
    return NULL;
}

static void float_to_short_array(const float *src, short *dest, int count)
{
    float scaled_value;
    int value;

    static const float normfact = 8.0 * 0x10000000;
    unsigned char *ucptr = ((unsigned char*) dest) + 2 * count;

    while (--count >= 0)
    {
	ucptr -= 2 ;
        scaled_value = src[count] * normfact;
        if (scaled_value >= (1.0 * 0x7FFFFFFF))
        {
            ucptr[0] = 0xFF;
            ucptr[1] = 0x7F;
            continue ;
        }
        if (scaled_value <= (-8.0 * 0x10000000))
        {
            ucptr[0] = 0x00;
            ucptr[1] = 0x80;
            continue;
        }

        value = lrintf(scaled_value);
        ucptr[0] = value >> 16;
        ucptr[1] = value >> 24;
    }
}

static int readfile(const char *filename, int *sample_rate, float **res,
    int *res_cnt)
{
    SF_INFO sfinfo;
    SNDFILE *infile;
    const int channels = 1; // TODO: add channel selection or mix
    float *data;
    int total = 0;
    int cnt;
    
    *res = NULL;
    *res_cnt = 0;
    
    // open file
    memset(&sfinfo, 0, sizeof(sfinfo));
    infile = sf_open(filename, SFM_READ, &sfinfo);
    if (infile == NULL)
    {
        return RESAMPLED_ERROR_FILE_READ;
    }
    
    // check format
    if (sfinfo.channels != 1)
    {
        sf_close(infile);
        return RESAMPLED_ERROR_FILE_FORMAT;
    }
    
    // allocate buffer
    data = malloc(sfinfo.frames * channels * sizeof(float));
    if (data == NULL)
    {
        sf_close(infile);
        return RESAMPLED_ERROR_MEMORY;
    }
    
    // read file
    while (total != sfinfo.frames)
    {
        cnt = sf_readf_float(infile, data + total * channels, sfinfo.frames / channels - total);
        if (cnt < 0)
        {
            free(data);
            sf_close(infile);
            return RESAMPLED_ERROR_FILE_READ;
        }
        total += cnt;
    }    
    sf_close(infile);
    
    *res = data;
    *res_cnt = sfinfo.frames;
    *sample_rate = sfinfo.samplerate;
    return RESAMPLED_ERROR_OK;
}

static int sample_rate_convert(int sample_rate, int new_sample_rate,
    float *data, int data_c, float **dst, int *dst_c)
{
    SRC_STATE *src_state;
    SRC_DATA src_data;
    int	error;
    float *output;
    float *ptr;
    int output_c;
    int output_a;
    double src_ratio;
    
    *dst = NULL;
    *dst_c = 0;
    
    // don't resample if format is the same
    if (sample_rate == new_sample_rate)
    {
        *dst = malloc(sizeof(float) * data_c);
        if (*dst == NULL)
        {
            return RESAMPLED_ERROR_MEMORY;
        }
        
        memcpy(*dst, data, sizeof(float) * data_c);
        *dst_c = data_c;
        return RESAMPLED_ERROR_OK;
    }

    src_ratio = (1.0 * new_sample_rate) / sample_rate;
    if (src_is_valid_ratio(src_ratio) == 0)
    {
        return RESAMPLED_ERROR_PROCESS;
    }

    src_state = src_new(SRC_SINC_MEDIUM_QUALITY, 1, &error);
    if (src_state == NULL)
    {
        return RESAMPLED_ERROR_PROCESS;
    }
    
    output = malloc(sizeof(*output) * BUFFER_LEN);
    if (output == NULL)
    {
        src_delete(src_state);
        return RESAMPLED_ERROR_MEMORY;
    }
    output_c = 0;
    output_a = BUFFER_LEN;

    src_data.data_in = data;
    src_data.input_frames = data_c;
    src_data.end_of_input = SF_TRUE;
    src_data.src_ratio = src_ratio;
    src_data.data_out = output;
    src_data.output_frames = BUFFER_LEN;
    
    while (1)
    {        
        error = src_process(src_state, &src_data);
        if (error)
        {
            src_delete(src_state);
            free(output);
            return RESAMPLED_ERROR_PROCESS;
        }

        /* Terminate if done. */
        if (src_data.output_frames_gen == 0)
            break;

        output_c += src_data.output_frames_gen;

        src_data.data_in += src_data.input_frames_used;
        src_data.input_frames -= src_data.input_frames_used;
        
        if ((output_a - output_c) < BUFFER_LEN)
        {
            ptr = realloc(output, sizeof(*output) * (output_a + BUFFER_LEN));
            if (ptr == NULL)
            {
                src_delete(src_state);
                free(output);
                return RESAMPLED_ERROR_MEMORY;
            }
            output = ptr;
            output_a += BUFFER_LEN;
        }
        
        src_data.data_out = output + output_c;
    }
    
    
    src_delete(src_state);    
    *dst = output;
    *dst_c = output_c;
    
    //printf("Sample Rate   : %d -> %d\n", sample_rate, new_sample_rate);
    //printf("Frames        : %d -> %d\n", data_c, (int) output_c);
    return RESAMPLED_ERROR_OK;
}

static int resample(resampled_array_t data, const char *filename)
{
    int i;
    int res;
    int sample_rate;        
    float *samples;
    int samples_c;
    float *output;
    int output_c;
    short *result;
    
    res = readfile(filename, &sample_rate, &samples, &samples_c);
    if (res)
        return res;
    
    for (i = 0; i < SRATE_COUNT; i++)
    {
        res = sample_rate_convert(sample_rate, srate_values[i],
            samples, samples_c, &output, &output_c);
        if (res)
        {
            free(samples);
            return res;
        }

        result = malloc(sizeof(short) * output_c);
        if (result == NULL)
        {
            free(samples);
            free(output);
            return RESAMPLED_ERROR_MEMORY;
        }

        float_to_short_array(output, result, output_c);
        free(output);
            
        data[i].samples = result;
        data[i].samples_c = output_c;
    }

    free(samples);    
    return RESAMPLED_ERROR_OK;
}

void resampled_free(resampled_array_t *arr)
{
    if (*arr == NULL)
        return;

    int i;
    for (i = 0; i < SRATE_COUNT; i++)
        resampled_data_free(&(*arr)[i]);
    free(*arr);
    *arr = NULL;
}

int resampled_init(resampled_array_t *arr, const char *filename)
{
    *arr = malloc(SRATE_COUNT * sizeof(struct resampled_data));
    if (*arr == NULL)
        return RESAMPLED_ERROR_MEMORY;
    
    int i;
    for (i = 0; i < SRATE_COUNT; i++)
        resampled_data_init(&(*arr)[i]);
    
    int res = resample(*arr, filename);
    if (res)
    {
        resampled_free(arr);
        return res;
    }
 
    return RESAMPLED_ERROR_OK;
}
